﻿local db, rosterCooldown, guildName, doUpdate, debugMode
local guildRoster = {}
local realm = GetRealmName()


local addon = CreateFrame("Frame")
addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("PLAYER_LOGIN")
addon:RegisterEvent("PLAYER_GUILD_UPDATE")
addon:SetScript("OnEvent", function(self, event, ...)
	return self[event](self, ...)
end)
addon:Hide() -- editboxes will, for some reason, be empty unless we hide this frame initially


do
	addon.name = ...
	InterfaceOptions_AddCategory(addon)
	
	local title = addon:CreateFontString(nil, nil, "GameFontNormalLarge")
	title:SetPoint("TOPLEFT", 16, -16)
	title:SetPoint("RIGHT", -16, 0)
	title:SetJustifyH("LEFT")
	title:SetJustifyV("TOP")
	title:SetText(addon.name)
	
	local function createButton()
		local btn = CreateFrame("CheckButton", nil, addon, "OptionsBaseCheckButtonTemplate")
		btn:SetPushedTextOffset(0, 0)

		local text = btn:CreateFontString(nil, nil, "GameFontHighlight")
		text:SetPoint("LEFT", btn, "RIGHT", 0, 1)
		btn:SetFontString(text)
		
		return btn
	end

	local detailsBtn = createButton()
	detailsBtn:SetPoint("TOPLEFT", title, "BOTTOMLEFT", -2, -16)
	detailsBtn:SetText("Detailed info")
	detailsBtn:SetScript("OnClick", function(self)
		local checked = self:GetChecked() == 1
		GuildWhoDB.details = checked
		PlaySound(checked and "igMainMenuOptionCheckBoxOn" or "igMainMenuOptionCheckBoxOff")
	end)
	addon.detailsBtn = detailsBtn

	local debugBtn = createButton()
	debugBtn:SetPoint("TOP", detailsBtn, "BOTTOM", 0, -8)
	debugBtn:SetText("Debug")
	debugBtn:SetScript("OnClick", function(self)
		local checked = self:GetChecked() == 1
		debugMode = checked
		PlaySound(checked and "igMainMenuOptionCheckBoxOn" or "igMainMenuOptionCheckBoxOff")
	end)

	local menu = {
		"None",
		"Public",
		"Officer",
	}

	local function onClick(self)
		UIDropDownMenu_SetSelectedValue(self.owner, self.value)
		db.note = self.value
		doUpdate = true
	end

	local dropdown = CreateFrame("Frame", "GuildWhoNoteChoice", addon, "UIDropDownMenuTemplate")
	UIDropDownMenu_SetWidth(dropdown, 120)
	dropdown:SetPoint("TOPLEFT", debugBtn, "BOTTOMLEFT", -15, -24)
	function dropdown:initialize()
		for _, v in ipairs(menu) do
			local info = UIDropDownMenu_CreateInfo()
			info.text = v
			info.func = onClick
			info.owner = self
			UIDropDownMenu_AddButton(info)
		end
	end
	addon.noteDropdown = dropdown

	local label = dropdown:CreateFontString(nil, "BACKGROUND", "GameFontNormalSmall")
	label:SetPoint("BOTTOMLEFT", dropdown, "TOPLEFT", 16, 3)
	label:SetText("Note:")

	local useCustom = createButton()
	useCustom:SetPoint("TOPLEFT", dropdown, "BOTTOMLEFT", 15, -8)
	useCustom:SetText("Use custom note")
	useCustom:SetScript("OnClick", function(self)
		local checked = self:GetChecked() == 1
		PlaySound(checked and "igMainMenuOptionCheckBoxOn" or "igMainMenuOptionCheckBoxOff")
		db.useCustom = checked
		doUpdate = true
	end)
	addon.useCustomBtn = useCustom
	
	local custom = CreateFrame("Editbox", "GuildWhoCustomEditbox", addon, "InputBoxTemplate")
	custom:SetWidth(128)
	custom:SetHeight(20)
	custom:SetPoint("TOPLEFT", useCustom, "BOTTOMLEFT", 9, -16)
	custom:SetAutoFocus(false)
	custom:SetScript("OnEnterPressed", function(self)
		self:ClearFocus()
		db.custom = self:GetText()
		doUpdate = true
	end)
	custom:SetScript("OnEscapePressed", custom.ClearFocus)
	addon.custom = custom
	
	local text = custom:CreateFontString(nil, nil, "GameFontNormalSmall")
	text:SetPoint("BOTTOMLEFT", custom, "TOPLEFT", 0, 2)
	text:SetText("Custom note")
	
	local ignore = CreateFrame("Editbox", "GuildWhoIgnoreEditbox", addon, "InputBoxTemplate")
	ignore:SetWidth(128)
	ignore:SetHeight(20)
	ignore:SetPoint("TOP", custom, "BOTTOM", 0, -20)
	ignore:SetAutoFocus(false)
	ignore:SetScript("OnEnterPressed", function(self)
		self:ClearFocus()
		db.ignore = self:GetText()
		doUpdate = true
	end)
	ignore:SetScript("OnEscapePressed", ignore.ClearFocus)
	addon.ignore = ignore
	
	local text = ignore:CreateFontString(nil, nil, "GameFontNormalSmall")
	text:SetPoint("BOTTOMLEFT", ignore, "TOPLEFT", 0, 2)
	text:SetText("Ignore pattern")
end


local function getToon(name)
	local character = guildRoster[name]
	if character then
		return character
	else
		for k, v in pairs(guildRoster) do
			local character = v.alts[name]
			if character then
				return character
			end
		end
	end
end

local function characterOnline(name)
	local toon = getToon(name)
	if toon then
		toon.online = true
		addon:Debug(name..".online =", true)
	end
end

-- someone went offline
local function characterOffline(name)
	local toon = getToon(name)
	if toon then
		toon.online = nil
		addon:Debug(name..".online =", nil)
	end
end

-- someone joins the guild, add to guild roster table
local function characterJoin(name)
	local character = guildRoster[name]
	if character then
		character.online = true
	end
	-- try to update guild roster to get data on the newcomer
	GuildRoster()
end

-- message matches a character leaving or being kicked out of the guild
local function characterLeave(name)
	local character = guildRoster[name]
	if character then
		-- keep this character as a "ghost" since it will have alts
		character.online = nil
		character.level = nil
		character.class = nil
		addon:Debug(name.." left guild -> keep as ghost.")
	else
		for k, v in pairs(guildRoster) do
			local alts = v.alts
			if alts[name] then
				alts[name] = nil
				addon:Debug(format("%s left guild and is alt -> delete from main %s.", name, k))
				-- remove this main if no alts remain
				if not next(alts) then
					guildRoster[k] = nil
					addon:Debug("Main has no remaining alts -> delete.")
				end
				break
			end
		end
	end
end

local chatMsgPatterns = {
	["^"..ERR_FRIEND_ONLINE_SS:format("([^|]+)", ""):gsub("%.", "%%."):gsub("%[%]", ".+").."$"] = characterOnline,
	["^"..ERR_FRIEND_OFFLINE_S:format("(%S+)"):gsub("%.", "%%.").."$"] = characterOffline,
	["^"..ERR_GUILD_JOIN_S:format("(%S+)"):gsub("%.", "%%.").."$"] = characterJoin,
	["^"..ERR_GUILD_LEAVE_S:format("(%S+)"):gsub("%.", "%%.").."$"] = characterLeave,
	["^"..ERR_GUILD_REMOVE_SS:format("(%S+)", "(%S+)"):gsub("%.", "%%.").."$"] = characterLeave,
}


local lastUpdate = 0

local function onUpdate(self, elapsed)
	lastUpdate = lastUpdate + elapsed
	if lastUpdate > 10 then
		self:Debug("Roster cooldown finished")
		rosterCooldown = false
		lastUpdate = 0
		self:SetScript("OnUpdate", nil)
	end
end


function addon:ADDON_LOADED(loadedAddon)
	if loadedAddon == "Blizzard_GuildUI" then
		local function onEnter(self)
			if doUpdate then
				addon:UpdateGuildRoster()
			end
			local name
			if UIDropDownMenu_GetSelectedValue(GuildRosterViewDropdown) == "tradeskill" then
				name = select(8, GetGuildTradeSkillInfo(self.guildIndex))
			else
				name = GetGuildRosterInfo(self.guildIndex)
			end
			local alts = addon:GetAlts(name, true)
			if alts then
				GameTooltip:SetOwner(self, "ANCHOR_RIGHT", 28, 0)
				GameTooltip:AddLine(name)
				for i, v in ipairs(alts) do
					if GuildWhoDB.details then
						local color = RAID_CLASS_COLORS[v.class]
						GameTooltip:AddDoubleLine(v.name, format(UNIT_LEVEL_TEMPLATE, v.level), color.r, color.g, color.b, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
					else
						GameTooltip:AddLine(v.name, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
					end
				end
				GameTooltip:Show()
			end
		end
		for i, v in ipairs(GuildRosterContainer.buttons) do
			v:HookScript("OnEnter", onEnter)
		end
		self:UnregisterEvent("ADDON_LOADED")
	end
end


function addon:Initialize()
	local newGuild = GetGuildInfo("player")
	
	if newGuild ~= guildName then
		self:Debug("Guild:", newGuild, GetNumGuildMembers())
	
		guildName = newGuild
		
		-- load/create saved variables if we are in a guild, and the guild name can be retrieved
		if guildName then
			self:Debug("Joined new guild.")
			
			local defaults = {
				note = "Public",
				useCustom = false,
				custom = "",
				ignore = "",
			}
			
			if GuildWhoDB then
				local realmData = GuildWhoDB[realm]
				if realmData then
					realmData[guildName] = realmData[guildName] or defaults
				else
					GuildWhoDB[realm] = {
						[guildName] = defaults
					}
				end
			else
				GuildWhoDB = {
					[realm] = {
						[guildName] = defaults
					}
				}
			end
			
			-- upgrade data
			for k, realm in pairs(GuildWhoDB) do
				if k ~= "details" then
					for k, guild in pairs(realm) do
						if type(guild) == "string" then
							realm[k] = {
								note = guild:gsub(".", string.upper, 1),
								useCustom = false,
								custom = "",
								ignore = "",
							}
						else
							break
						end
					end
				end
			end
			
			db = GuildWhoDB[realm][guildName]
			
			local dropdown = self.noteDropdown
			UIDropDownMenu_SetSelectedValue(dropdown, db.note)
			UIDropDownMenu_SetText(dropdown, db.note)
			self.useCustomBtn:SetChecked(db.useCustom)
			self.custom:SetText(db.custom)
			self.ignore:SetText(db.ignore)
			
			doUpdate = true
			GuildRoster()
			
			self:RegisterEvent("GUILD_ROSTER_UPDATE")
			self:RegisterEvent("CHAT_MSG_SYSTEM")
		else
			self:Debug("Left guild.")
			
			self:UnregisterEvent("GUILD_ROSTER_UPDATE")
			self:UnregisterEvent("CHAT_MSG_SYSTEM")
		end
	end
end


function addon:PLAYER_GUILD_UPDATE(unit)
	if unit == "player" then
		self:Initialize()
	end
end


function addon:PLAYER_LOGIN()
	GuildWhoDB = GuildWhoData or GuildWhoDB
	self:Initialize()
	GuildWhoDB.details = GuildWhoDB.details or false
	self.detailsBtn:SetChecked(GuildWhoDB.details)
end


function addon:CHAT_MSG_SYSTEM(msg)
	if not guildRoster then
		self:Debug("Guild roster cache not yet initialized -> return.")
		return
	end
	
	-- match the CHAT_MSG_SYSTEM message against global strings
	for pattern, func in pairs(chatMsgPatterns) do
		local name = msg:match(pattern)
		if name then
			func(name)
			break
		end
	end
end


function addon:GUILD_ROSTER_UPDATE()
	-- if forceUpdate then
		-- if guildName and GetNumGuildMembers() > 0 then
			-- self:UpdateGuildRoster()
			-- forceUpdate = nil
		-- end
	-- end
	if self.delayedCall then
		-- guild roster update was available for this call, so update local guild roster first
		self:UpdateGuildRoster()
		self:Debug("Delayed call")
		self[self.delayedCall](self, self.delayedCallArg)
		self.delayedCall = nil
	end
	if not rosterCooldown then
		-- calling GuildRoster has a 10 sec "cooldown"
		self:SetScript("OnUpdate", onUpdate)
		self:Debug("Roster cooldown start")
		rosterCooldown = true
	end
	doUpdate = true
end


function addon:DelayCall(method, name)
	self:Debug(" ----- ")
	
	if not guildName then
		self:Debug("Not in a guild - cancel call.")
		return true
	end
	
	if not (rosterCooldown or self.delayedCall) then
		-- roster cooldown is ready - delaying call
		self:Debug("Roster cooldown is ready - delay call.")
		self.delayedCall = method
		self.delayedCallArg = name
		GuildRoster()
		return true
	end
	-- roster on cooldown or a delayed call, get data now
	self:Debug("Roster on cooldown - get data now.")
end


function addon:GetMain(name)
	self:Debug(" ----- ")
	self:Debug("Querying "..name.."...")
	
	-- initiate custom guild roster now, should it not be ready
	local main = guildRoster[name]
	if main then
		self:Debug("mainChar = "..name)
		return name
	else
		for character, data in pairs(guildRoster) do
			data = data.alts[name]
			-- return the main's name, if it has an alt with the specified name
			if data then
				self:Debug("mainChar = "..character)
				return character
			end
		end
	end
	
	self:Debug("Found no data.")
end


function addon:FindOnline(name)
	if self:DelayCall("FindOnline", name) then
		return
	end
	
	local altChar
	local mainChar = self:GetMain(name)
	
	if not mainChar then
		-- character is not in guild, or has no main or alts
		self:Print("No data found on %s.", name)
		return
	end
	
	local main = guildRoster[mainChar]
	if main.online then
		altChar = mainChar
	else
		for k, v in pairs(main.alts) do
			if v.online then
				altChar = k
				break
			end
		end
	end
	
	self:Debug("altChar =", altChar)
	
	-- print result
	if altChar then
		if name ~= altChar then
			self:Print("%s is online on |Hplayer:%s|h[%s]|h.", name, altChar, altChar)
		else
			self:Print("|Hplayer:%s|h[%s]|h is online.", name, name)
		end
	else
		self:Print("%s is offline.", name)
	end
end


local function detailedSort(a, b)
	if a.level == b.level then
		return a.name < b.name
	else
		return a.level > b.level
	end
end

function addon:GetAlts(name)
	-- get main character of the given character
	local mainChar = self:GetMain(name)
	
	if not mainChar then
		return
	end
	
	local alts = {}
	local main = guildRoster[mainChar]
	
	-- include the main unless that's the one we used to search
	if mainChar ~= name and main.class then
		alts[#alts + 1] = {
			name = mainChar,
			online = main.online,
			level = main.level,
			class = main.class,
		}
	end
	
	for character, data in pairs(main.alts) do
		-- include the main's alts except the one we used to search
		if character ~= name then
			tinsert(alts, {
				name = character,
				online = data.online,
				level = data.level,
				class = data.class,
			})
		end
	end
	
	-- do not return any results if the main is a ghost and there are no other alts
	if #alts == 0 then
		self:Debug("Main is a ghost and no other alts exist.")
		return
	end
	
	-- sort alts by level -> name
	sort(alts, not GuildWhoDB.details and nil or detailedSort)
	
	return alts
end


function addon:PrintAlts(name)
	if self:DelayCall("PrintAlts", name) then
		return
	end
	
	local alts = self:GetAlts(name)
	
	if not alts then
		-- searched character wasn't a guild match or had no other characters
		self:Print("No data found on %s.", name)
		return
	end
	
	self:Print("Known alts of %s:", name)
	for i, v in ipairs(alts) do
		local name = v.name
		local online = ""
		local info = ""
		if GuildWhoDB.details then
			local color = RAID_CLASS_COLORS[v.class]
			name = format("|cff%2x%2x%2x%s|r", color.r * 255, color.g * 255, color.b * 255, name)
			info = format(" - %s %d", LEVEL, v.level)
		end
		if v.online then
			name = format("|Hplayer:%s|h[%s]|h", v.name, name)
			online = format(" (%s)", GUILD_ONLINE_LABEL)
		end
		self:Print("  %s%s%s", name, online, info)
	end
end


function addon:UpdateGuildRoster()
	local numGuildies = GetNumGuildMembers()
	
	doUpdate = nil
	
	if guildName and numGuildies > 0 then
		self:Debug(numGuildies.. " guild members.")
	else
		self:Debug("Not in a guild or guild roster not yet initialized - do not update.")
		return
	end
	
	self:Debug("Updating guild roster cache...")
	
	wipe(guildRoster)
	
	local note = db.note
	
	if note == "None" then
		return
	end
	
	local useOfficer = note == "Officer"
	local pattern = db.useCustom and format("^%s$", db.custom:trim()):gsub("[()%.[%]*+-?)]", "%%%1"):gsub("$n", "([^%%s%%d%%p]+)") or "([^%s%d%p]+)"
	local ignore = db.ignore
	
	for i = 1, numGuildies do
		local name, _, _, level, _, _, publicnote, officernote, online, _, classFile = GetGuildRosterInfo(i)
		if not name then break end
		local note = (useOfficer and officernote or publicnote):match(pattern)
		local guildMember = guildRoster[name] or {
			online = online,
			level = level,
			class = classFile,
		}
		-- if this character has a valid guild note it will be the name of the main, and so we create an entry for that "main"
		-- it also cannot be its own main
		if note and note ~= name and not note:match("^"..ignore.."$") then
			note = note:lower():gsub(".", string.upper, 1)
			local main = guildRoster[note] or {alts = {}}
			guildRoster[note] = main
			main.alts = main.alts or {}
			-- and insert the character in question as an alt of this main
			main.alts[name] = guildMember
		end
	end
	
	-- fill in data for mains as they were only created with a name
	for i = 1, numGuildies do
		local name, _, _, level, _, _, _, _, online, _, classFile = GetGuildRosterInfo(i)
		if not name then break end
		local guildMember = guildRoster[name]
		if guildMember then
			guildMember.online = online
			guildMember.level = level
			guildMember.class = classFile
		end
	end
	
	return true
end


function addon:Print(msgFormat, ...)
	DEFAULT_CHAT_FRAME:AddMessage(format(msgFormat, ...), 1, 1, 0)
end


-- debug message function, only visible in debug mode
function addon:Debug(...)
	if debugMode then
		print("|cffffff00•", ...)
	end
end


function addon:PrintHelp()
	self:Print("- How to use GuildWho:")
	self:Print("/alts character - list this character's alts.")
	self:Print("/gwho character - find who this character is online on.")
end


SlashCmdList["GWHO_WHO"] = function(msg)
	if msg == "" then
		addon:PrintHelp()
	else
		addon:FindOnline(msg:lower():gsub(".", string.upper, 1))
	end
end
SLASH_GWHO_WHO1 = "/gwho"

SlashCmdList["GWHO_ALTS"] = function(msg)
	if msg == "" then
		self:PrintHelp()
	else
		addon:PrintAlts(msg:lower():gsub(".", string.upper, 1))
	end
end
SLASH_GWHO_ALTS1 = "/alts"


hooksecurefunc("UnitPopup_OnClick", function(self)
	local name = UIDROPDOWNMENU_INIT_MENU.name
	
	if name and self.value == "GW_LISTALTS" then
		addon:PrintAlts(name)
	end
end)


hooksecurefunc("UnitPopup_HideButtons", function()
	local dropdownMenu = UIDROPDOWNMENU_INIT_MENU
	
	for index, value in ipairs(UnitPopupMenus[dropdownMenu.which]) do
		if value == "GW_LISTALTS" then
			-- only show this dropdown menu entry if the character is in the guild
			UnitPopupShown[UIDROPDOWNMENU_MENU_LEVEL][index] = addon:UpdateGuildRoster() and addon:GetMain(dropdownMenu.name) and 1 or 0
			break
		end
	end
end)

-- add a "List alts" entry to unit popup menus
UnitPopupButtons["GW_LISTALTS"] = {text = "List alts", dist = 0}
tinsert(UnitPopupMenus["FRIEND"], #UnitPopupMenus["FRIEND"] - 1, "GW_LISTALTS")